from .focalloss import FocalLoss
from .soft_cross_entropy_loss import SoftCrossEntropyLoss
from .swem import SWEM
from .textcnn import TextCNN

__all__ = [
    "FocalLoss",
    "SWEM",
    "SoftCrossEntropyLoss",
    "TextCNN",
]
__dir__ = lambda: __all__



# references
# https://kornia.readthedocs.io/en/latest/_modules/kornia/losses/focal.html


import torch
import torch.nn as nn
import torch.nn.functional as F


class FocalLoss(nn.Module):
    """Focal Loss

    - https://arxiv.org/abs/1708.02002/

    Args:
        num_classes (int):
        alpha (float):
        gamma (float):
        reduction (str):
    """

    def __init__(
        self,
        num_classes: int,
        alpha: float,
        gamma: float = 2.0,
        reduction: str = "none",
    ) -> None:
        super(FocalLoss, self).__init__()

        self.num_classes: int = num_classes
        self.alpha: float = alpha
        self.gamma: float = gamma
        self.reduction: str = reduction

    def forward(
        self,
        y_pred: torch.Tensor,
        y_true: torch.Tensor,
    ) -> torch.Tensor:
        """TODO

        Args:
            y_pred (torch.Tensor):
            y_true (torch.Tensor):

        Returns:
            loss (torch.Tensor):
        """
        # convert y_true to one-hot encoding
        y_true_one_hot = F.one_hot(y_true, self.num_classes).type(y_pred.dtype)

        # softmax(y_pred) and log(softmax(y_pred))
        softmax_y_pred = F.softmax(y_pred, dim=-1)
        log_softmax_y_pred = F.log_softmax(y_pred, dim=-1)

        focal = -self.alpha * torch.pow(1.0 - softmax_y_pred, self.gamma) * log_softmax_y_pred
        loss = torch.einsum("ab...,ab...->a...", (y_true_one_hot, focal))

        match self.reduction:
            case "none":
                return loss
            case "mean":
                return torch.mean(loss)
            case "sum":
                return torch.sum(loss)
            case _:
                raise ValueError



from dataclasses import dataclass

import numpy as np
from sklearn.metrics import accuracy_score, precision_recall_fscore_support


@dataclass
class APRFSScores:
    # accuracy
    a: float
    # none
    p: list[float]
    r: list[float]
    f: list[float]
    s: list[float]
    # macro
    macro_p: float
    macro_r: float
    macro_f: float
    # mecro
    micro_p: float
    micro_r: float
    micro_f: float


def accuracy_precision_recall_fscore_support(
    y_true: np.ndarray,
    y_pred: np.ndarray,
):
    """TODO

    Args:
        y_true (np.ndarray):
        y_pred (np.ndarray):

    Returns:
        aprfs_scores (APRFSScores):
    """

    a = accuracy_score(y_true=y_true, y_pred=y_pred)
    p, r, f, s = precision_recall_fscore_support(y_true=y_true, y_pred=y_pred)
    macro_p, macro_r, macro_f, _ = precision_recall_fscore_support(
        y_true=y_true, y_pred=y_pred, average="macro"
    )
    micro_p, micro_r, micro_f, _ = precision_recall_fscore_support(
        y_true=y_true, y_pred=y_pred, average="micro"
    )

    aprfs_scores = APRFSScores(a, p, r, f, s, macro_p, macro_r, macro_f, micro_p, micro_r, micro_f)

    return aprfs_scores



import torch.nn as nn
import torch.nn.functional as F


class SoftCrossEntropyLoss(nn.Module):
    def __init__(self, weights=None, reduction="mean"):
        """TODO

        Args:
            weights (torch.Tensor, optional): TODO
            reduction (str, optional): TODO
        """
        super(SoftCrossEntropyLoss, self).__init__()

        self.weights = weights
        self.reduction = reduction

    def forward(self, y_logits, y_target):
        """TODO

        Args:
            y_logits (torch.Tensor): TODO
            y_target (torch.Tensor): TODO

        Returns:
            loss (torch.Tensor)
        """
        logsm_y_pred = F.log_softmax(y_logits, dim=1)

        if self.weights is not None:
            y_target = self.weights * y_target

        loss = -(logsm_y_pred * y_target).sum() / (y_target).sum()

        return loss



from .cat_and_zero_pad_arrays_with_different_shape_zero import (
    cat_and_zero_pad_arrays_with_different_shape_zero,
)

__all__ = [
    "cat_and_zero_pad_arrays_with_different_shape_zero",
]
__dir__ = lambda: __all__



# reference
# https://christopher5106.github.io/deep/learning/2020/04/02/fasttext_pretrained_embeddings_subword_word_representations.html

import fasttext
import numpy as np


class FastTextExtension:
    """FastTextExtension

    Args:
        filename (str): /path/to/bin/file
    """

    def __init__(self, filename: str) -> None:
        self.ft: fasttext.FastText._FastText = fasttext.load_model(filename)

    @property
    def embeddings(self) -> np.ndarray:
        """Get word and subword embeddings.

        Returns:
            embeddings (np.ndarray): word and subword embeddings
        """
        return self.ft.get_input_matrix()

    @property
    def vocab(self) -> list[str]:
        """Get vocabulary list.

        Returns:
            vocab (list[str]): list of vocabularies
        """
        return self.ft.get_words()

    def get_word_vectors(self, sent: str, norm: bool = False) -> list[np.ndarray]:
        """Get list of word vectors.

        Args:
            sent (str): sentence
            norm (bool): normalize each word vectors

        Returns:
            vectors (list[np.ndarray]): list of word vectors
        """
        words: list[str] = fasttext.tokenize(sent)
        vectors: list[np.ndarray] = []

        for word in words:
            vector = self.ft.get_word_vector(word)

            if norm:
                vector_norm = np.linalg.norm(vector)

                if vector_norm > 0:
                    vector /= vector_norm

            vectors.append(vector)

        return vectors



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F


class TextCNN(nn.Module):
    """TextCNN

    Args:
        num_embeddings (int):
        embedding_dim (int):
        pretrained_embedding (Optional[torch.Tensor]):
        conv_kernel_sizes (list[int]):
        conv_out_channels (int):
        dropout_p (float):
        cls_hidden_size (int):
        cls_num_classes (int):
        freeze_pretrained_embeddings (bool):
    """

    def __init__(
        self,
        # embeddings
        num_embeddings: int,
        embedding_dim: int,
        pretrained_embedding: Optional[torch.Tensor],
        # convs
        conv_kernel_sizes: list[int],
        conv_out_channels: int = 128,
        # dropout
        dropout_p: float = 0.5,
        # classifier
        cls_hidden_size: int = 256,
        cls_num_classes: int = 1,
        #
        freeze_pretrained_embeddings: bool = False,
    ) -> None:
        super(TextCNN, self).__init__()

        # embeddings
        self.embedding = nn.Embedding(num_embeddings=num_embeddings, embedding_dim=embedding_dim)

        # load pretrained embeddings if provided
        if pretrained_embedding is not None:
            self.embedding.from_pretrained(
                pretrained_embedding, freeze=freeze_pretrained_embeddings
            )

        # convs
        self.convs = nn.ModuleList(
            [
                nn.Conv2d(
                    in_channels=1,
                    out_channels=conv_out_channels,
                    kernel_size=(kernel_size, embedding_dim),
                )
                for kernel_size in conv_kernel_sizes
            ]
        )

        # dropout
        self.dropout = nn.Dropout(dropout_p)

        # classifier
        self.classifier = nn.Sequential(
            nn.Linear(conv_out_channels * len(conv_kernel_sizes), cls_hidden_size),
            nn.ReLU(),
            nn.Linear(cls_hidden_size, cls_num_classes),
        )

    def forward(self, x: torch.LongTensor) -> torch.FloatTensor:
        """TODO

        Args:
            x (torch.LongTensor): TODO

        Returns:
            y_logits (torch.FloatTensor): TODO

        """
        # embeddings
        z_emb = self.embedding(x)  # shape(batch_size, max_length, embedding_dim)
        # (batch_size, in_channels=1, max_length, embedding_dim)
        z_emb_unsqueezed = z_emb.unsqueeze(1)  # add depth channel

        # convs
        # shape[(bs, out_channels, max_length - kernel_size + 1)]
        z_convs = [conv(z_emb_unsqueezed).squeeze() for conv in self.convs]

        # pooling
        # shape[(bs, out_channels, 1)]
        z_pooled = [F.max_pool1d(z_conv, z_conv.shape[2]) for z_conv in z_convs]

        # shape(bs, out_channels * len(kernel_sizes))
        z_cat = torch.cat(z_pooled, dim=1).squeeze(-1)

        # dropout
        z_dropout = self.dropout(z_cat)

        # classifier
        y_logits = self.classifier(z_dropout)

        return y_logits



import numpy as np


def cat_and_zero_pad_arrays_with_different_shape_zero(tensors: list[np.ndarray]) -> np.ndarray:
    """

    Args:
        tensors (list[np.ndarray]):

    Returns:
    """
    max_length = 0

    for tensor in tensors:
        max_length = max(max_length, tensor.shape[0])

    cat = np.zeros((len(tensors), max_length, *tensors[0].shape[1:]))

    for i, tensor in enumerate(tensors):
        cat[i, : tensor.shape[0]] += tensor

    return cat
